//   Copyright (c)  1997-2006  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License (version 2)
//   as published by the Free Software Foundation.  A copy of the full
//   licence may be found in the file COPYING in this directory.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoA; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <stdio.h>
#include "Zdet.h"

#include <math.h>
#include "mpz_log.h"
#include "jalloc.h"
#include "FF.h"
#include "FFdet.h"
#include "mpz_cra_ui.h"
#include "primes.h"
#include "add_logs.h"
#include "Zsolve.h"

/***************************************************************************/
static void Zmat_to_FFmat(FFelem **Mp, mpz_t **M, int r, int s)
{
  int i, j, p;

  p = CurrentFF.prime;
  for (i=0; i < r; i++)
    for (j=0; j < s; j++)
      Mp[i][j] = mpz_fdiv_ui(M[i][j], p);
}

/*******************************************************************/
/* This computes the log (approx) of Hadamard's determinant bound. */
/* We compute both for rows and columns then take the smaller.     */

static double Zmat_det_bound(mpz_t **M, int n)
{
  int i, j;
  double this_row, this_col;
  double row_bd, col_bd;

  row_bd = 0;
  for (i=0; i < n; i++)
  {
    this_row = -99; /* log of "zero" */
    for (j=0; j < n; j++)
      if (mpz_sgn(M[i][j])) this_row = add_logs(this_row, 2*mpz_log(M[i][j]));
    row_bd += this_row/2;
  }

  col_bd = 0;
  for (i=0; i < n; i++)
  {
    this_col = -99; /* log of "zero" */
    for (j=0; j < n; j++)
      if (mpz_sgn(M[j][i])) this_col = add_logs(this_col, 2*mpz_log(M[j][i]));
    col_bd += this_col/2;
  }

  if (col_bd < row_bd) return col_bd;
  return row_bd;
}



/***************************************************************************/

void Zdet(mpz_t det, int n, mpz_t **M)
{
  mpz_t **rhs, **soln, *denom, D;
  int i, j, det_changed, r, singular;
  FFelem p, detp, **Mp, Dp;
  double log_det_modulus, det_bound;
  mpz_t det_modulus;
  FF FFp;
  PrimeSource PS = PrimeSourceCtor();

  r = 1;
  rhs = (mpz_t **)MALLOC(n*sizeof(mpz_t *));
  soln = (mpz_t **)MALLOC(n*sizeof(mpz_t *));
  denom = (mpz_t*)MALLOC(r*sizeof(mpz_t));
  for(j=0; j < r; j++) mpz_init(denom[j]);
  for(i=0; i < n; i++)
  {
    rhs[i] = (mpz_t*)MALLOC(r*sizeof(mpz_t));
    soln[i] = (mpz_t*)MALLOC(r*sizeof(mpz_t));
    for(j=0; j < r; j++)
    {
      mpz_init(rhs[i][j]);
      mpz_init(soln[i][j]);
    }
  }
  for (i=0; i < n; i++) mpz_set_si(rhs[i][0], (random()%199)-99);

  singular = Zsolve(soln, denom, n, M, r, rhs);
  if (singular == 0) printf("System was singular.\n");

  mpz_init_set(D, denom[0]); /* ASSUMES r = 1 */
  /* free rhs, soln, and denom */
  for(j=0; j < r; j++) mpz_clear(denom[j]);
  FREE(denom);
  for(i=0; i < n; i++)
  {
    for(j=0; j < r; j++)
    {
      mpz_clear(rhs[i][j]);
      mpz_clear(soln[i][j]);
    }
    FREE(rhs[i]);
    FREE(soln[i]);
  }
  FREE(rhs);
  FREE(soln);

  mpz_set_ui(det, 0);
  mpz_init_set_ui(det_modulus, 1);
  log_det_modulus = 0;
  Mp = (FFelem**)MALLOC(n*sizeof(FFelem*));
  for (i=0; i < n; i++) Mp[i] = (FFelem*)MALLOC(n*sizeof(FFelem));

  det_bound = 1 + Zmat_det_bound(M, n); /* add 1 to allow for approx error */
printf("Hadamard bound in bits is %f\n", det_bound/log(2.0));
  det_bound -= mpz_log(D);
printf("Remaining distance to lift: %f\n", det_bound/log(2.0));
print_time();
  while (log_det_modulus < det_bound)
  {
    p = NextPrime(PS);
    Dp = mpz_fdiv_ui(D, p);
    if (Dp == 0) continue;
    FFp = FFctor(p);
    FFselect(FFp);
    Zmat_to_FFmat(Mp, M, n, n);
    detp = FFdet(n, Mp);
    detp = FFdiv(detp, Dp);
    det_changed = mpz_cra_ui(det, det_modulus, detp, p);
    mpz_mul_ui(det_modulus, det_modulus, p);
    log_det_modulus += log(p);
    FFdtor(FFp);
  }
  mpz_clear(det_modulus);
  PrimeSourceDtor(PS);
  for (i=0; i<n; i++) FREE(Mp[i]);
  FREE(Mp);
  mpz_mul(det, det, D);
  mpz_clear(D);
}


